# --------------------------------------------------------
# (c) Copyright 2014, 2020 by Jason DeLaat.
# Licensed under BSD 3-clause licence.
# --------------------------------------------------------
""" Adds operators to the List monad. """
from typing import List, TypeVar

import pymonad.list
import pymonad.monad
import pymonad.operators.operators

T = TypeVar('T') # pylint: disable=invalid-name

class _List(pymonad.operators.operators.MonadOperators, pymonad.list._List[T]): # pylint: disable=protected-access
    """ See pymonad.operators.operators and pymonad.list. """

def ListMonad(*elements: List[T]) -> _List[T]: # pylint: disable=invalid-name
    """ Creates an instance of the List monad.

    Args:
      *elements: any number of elements to be inserted into the list

    Returns:
      An instance of the List monad.
    """

    return _List(list(elements), None)

ListMonad.apply = _List.apply
ListMonad.insert = _List.insert



# --------------------------------------------------------
# (c) Copyright 2014, 2020 by Jason DeLaat.
# Licensed under BSD 3-clause licence.
# --------------------------------------------------------
""" Provides monad classes with operators.

The operators module overrides all base monad types and their aliases
with versions which support operators.

The defined operators are:
  &  - amap
  >> - bind
  << - map
  *  - then

  Example:
    from pymonad.operators import Maybe, Just, Nothing
    from pymonad.tools import curry

    @curry(2)
    def add(x, y): return x + y

    @curry(2)
    def div(y, x):
        if y == 0:
            return Nothing
        else:
            return Just(x / y)

    # Equivalent to Maybe.apply(add).to_arguments(Just(1), Just(2))
    print(add << Just(1) & Just(2)) # Just 3

    # Equivalent to Maybe.insert(5).bind(div(2)).bind(div(0))
    print(Just(5) >> div(2) >> div(0)) # Nothing

    # Alternatively, each could be written with *
    print(add * Just(1) & Just(2)) # Just 3
    print(Just(5) * div(2) * div(0)) # Nothing
"""



# --------------------------------------------------------
# (c) Copyright 2014, 2020 by Jason DeLaat.
# Licensed under BSD 3-clause licence.
# --------------------------------------------------------
""" Implements the List monad.

The List monad is frequently used to represent calculations with
non-deterministic results, that is: functions which return more than
one (possible) result. For example, calculating how chess pieces might
move.

  Example:
    def knight_move(position):
        # calculates a list of every possible square a knight could move
        # to from it's current position
        return ListMonad(position_1, position_2, ..., position_N)

    # A list containing every square a knight could reach after 3 moves.
    three_moves = (List
                   .insert(initial_position) # However positions are defined.
                   .then(knight_move)
                   .then(knight_move)
                   .then(knight_move))
"""
from typing import Any, Callable, Generic, List, TypeVar, Union # pylint: disable=unused-import

import pymonad.monad

S = TypeVar('S') # pylint: disable=invalid-name
T = TypeVar('T') # pylint: disable=invalid-name

class _List(pymonad.monad.Monad, Generic[T]):
    @classmethod
    def insert(cls, value: T) -> '_List[T]':
        return ListMonad(value)

    def amap(self: '_List[Callable[[S], T]]', monad_value: '_List[S]') -> '_List[T]':
        result = []
        for function in self:
            for value in monad_value:
                result.append(function(value))
        return ListMonad(*result)

    def bind(self: '_List[S]', kleisli_function: Callable[[S], '_List[T]']) -> '_List[T]':
        return self.map(kleisli_function).join()

    def join(self: '_List[_List[T]]') -> '_List[T]':
        """ Flattens a nested ListMonad instance one level. """
        return ListMonad(*[element for lists in self for element in lists])

    def map(self: '_List[S]', function: Callable[[S], T]) -> '_List[T]':
        return ListMonad(*[function(x) for x in self])

    def then(
            self: '_List[S]', function: Union[Callable[[S], T], Callable[[S], '_List[T]']]
    ) -> '_List[T]':
        try:
            return self.bind(function)
        except TypeError:
            return self.map(function)

    def __eq__(self, other):
        return self.value == other.value

    def __getitem__(self, index):
        try:
            return ListMonad(*self.value.__getitem__(index))
        except TypeError:
            return self.value[index]

    def __iter__(self):
        return iter(self.value)

    def __len__(self):
        return len(self.value)

    def __repr__(self):
        return str(self.value)

def ListMonad(*elements: List[T]) -> _List[T]: # pylint: disable=invalid-name
    """ Creates an instance of the List monad.

    Args:
      *elements: any number of elements to be inserted into the list

    Returns:
      An instance of the List monad.
    """

    return _List(list(elements), None)

ListMonad.insert = _List.insert
ListMonad.apply = _List.apply



# --------------------------------------------------------
# (c) Copyright 2014, 2020 by Jason DeLaat.
# Licensed under BSD 3-clause licence.
# --------------------------------------------------------
""" Implements the Reader monad.

The Reader monad creates a context in which functions have access to
an additional read-only input.
"""
from typing import Any, Callable, Generic, TypeVar, Union

import pymonad.monad
import pymonad.tools

R = TypeVar('R') # pylint: disable=invalid-name
S = TypeVar('S') # pylint: disable=invalid-name
T = TypeVar('T') # pylint: disable=invalid-name

@pymonad.tools.curry(3)
def _bind(function_f, function_g, read_only):
    return function_f(function_g(read_only))(read_only)

@pymonad.tools.curry(3)
def _bind_or_map(function_f, function_g, read_only):
    try:
        return _bind(function_f, function_g, read_only)
    except TypeError:
        return _map(function_f, function_g, read_only)

@pymonad.tools.curry(3)
def _map(function_f, function_g, read_only):
    return function_f(function_g(read_only))

class _Reader(pymonad.monad.Monad, Generic[R, T]):
    @classmethod
    def insert(cls, value: T) -> '_Reader[Any, T]':
        return Reader(lambda r: value)

    def amap(self: '_Reader[R, Callable[[S], T]]', monad_value: '_Reader[R, S]') -> '_Reader[R, T]':
        return Reader(lambda r: self(r)(monad_value(r)))

    def bind(
            self: '_Reader[R, S]', kleisli_function: Callable[[S], '_Reader[R, T]']
    ) -> '_Reader[R, T]':
        return Reader(_bind(kleisli_function, self)) # pylint: disable=no-value-for-parameter

    def map(self: '_Reader[R, S]', function: Callable[[S], T]) -> '_Reader[R, T]':
        return Reader(_map(function, self)) # pylint: disable=no-value-for-parameter

    def then(
            self: '_Reader[R, S]', function: Union[Callable[[S], T], Callable[[S], '_Reader[R, T]']]
    ) -> '_Reader[R, T]':
        return Reader(_bind_or_map(function, self)) # pylint: disable=no-value-for-parameter

    def __call__(self, arg: R) -> T:
        return self.value(arg)

def Reader(function: Callable[[R], T]) -> _Reader[R, T]: # pylint: disable=invalid-name
    """ Creates an instance of the Reader monad.

    Args:
      function: a function which takes the read-only data as input and
        returns any appropriate type.

    Result:
      An instance of the Reader monad.
    """
    return _Reader(function, None)

Reader.apply = _Reader.apply
Reader.insert = _Reader.insert





def Compose(function: Callable[[R], T]) -> _Reader[R, T]: # pylint: disable=invalid-name
    """ Creates an instance of the Compose monad.

    Compose is basically an alias for the Reader monad except with the
    insert and apply methods removed. It's purpose is simply to
    provide a semantically meaningful monad instance to be used
    specifically for the purpose of function composition.

      Example:
        def inc(x): return x + 1
        def dec(x): return x - 1

        convoluted_inc_twice = (Compose(inc)
                                .then(inc)
                                .then(inc)
                                .then(dec))

        convoluted_inc_twice(0) # Result: 2

    Technically, 'convoluted_inc_twice' is an instance of the Reader
    monad but since Reader defines the __call__ method, we can treat
    it just like a function for all intents and purposes. The Compose
    monad composes functions forward. In the example, the three 'inc'
    operations happen first and then the 'dec' and not vice-versa.
    """
    return _Reader(function, None)






class _Pipe(pymonad.monad.MonadAlias, _Reader[R, T]):
    def flush(self):
        """ Calls the composed Pipe function returning  the embedded result.

        The 'flush' method calls the composed function with dummy
        input since all functions in a Pipe chain should ignore that
        input anyway, simply joining inputs to outputs.
        """
        return self(None)

    def __pos__(self):
        return self.flush()

def Pipe(value: T) -> _Pipe[Any, T]: # pylint: disable=invalid-name
    """ Creates an instance of the Pipe monad.

    Pipe is basically an alias for the Reader monad except with the
    insert and apply methods removed. It's purpose is simply to
    provide a semantically meaningful monad instance to be used
    specifically for the purpose of chaining function calls by taking
    the output of one function as the input to the next.

    Since Pipe is a subclass of Reader it's really building a function
    but, semantically, pipes start with some input and end with a
    result. For this reason, Pipe adds a 'flush' method which calls
    the composed function with dummy input (which will be ignored) and
    simply returns the embedded result. Optionally, you can instead
    use the unary '+' operator instead of 'flush' to do the same
    thing.

      Example:
        def inc(x): return x + 1

        pipe_with_flush = (Pipe(0)
                          .then(inc)
                          .then(inc)
                          .flush())

        pipe_with_plus = +(Pipe(0)
                           .then(inc)
                           .then(inc))

        pipe_with_flush == pipe_with_plus # True
    """
    return _Pipe.insert(value)



# --------------------------------------------------------
# (c) Copyright 2020 by Jason DeLaat.
# Licensed under BSD 3-clause licence.
# --------------------------------------------------------
""" The tools module contains useful functions that don't really belong anywhere else. """

from typing import Any, Callable, List, TypeVar

import pymonad.monad as monad

R = TypeVar('R') # pylint: disable=invalid-name
S = TypeVar('S') # pylint: disable=invalid-name
T = TypeVar('T') # pylint: disable=invalid-name

def _curry_helper(
        number_of_arguments: int, function_to_curry: Callable, accumulated_arguments: List[Any]
) -> Callable:
    """ Builds a curried version of the supplied function and returns it to the caller.

    Args:
      number_of_arguments: specifies how many arguments 'function_to_curry' takes as input.
      function_to_curry: a function or other callable object either built-in or user defined.
      accumulated_arguments: a list containing all the arguments so
        far supplied to 'function_to_curry'. For instance, if
        'function_to_curry' takes two arguments but has been called with
        only one so far, then accumulated_arguments will contain a
        single item.

    Returns:
      A new function which may be partially applied simply by passing
      the desired number of arguments.
    """
    def _curry_internal(*arguments: List[Any]):
        """ Handles the actual partial application of curried functions.

        Args:
          *arguments: a variable number of arguments to be
             (eventually) supplied to the wrapped function
             'function_to_curry'

        Returns:
          Either the result of calling the wrapped function_to_curry
          with the total accumulated arguments or the result of
          calling _curry_helper with the new, larger
          accumulated_arguments list.
        """
        all_arguments = accumulated_arguments[:]
        all_arguments.extend(arguments)
        if len(all_arguments) >= number_of_arguments: # pylint: disable=no-else-return
            return function_to_curry(*all_arguments)
        else:
            return _curry_helper(number_of_arguments, function_to_curry, all_arguments)
    return _curry_internal

# Use _curry_helper to define curry as itself being a curried function.
curry = _curry_helper(2, lambda n, f: _curry_helper(n, f, []), []) # pylint: disable=invalid-name
curry.__doc__ = """ Creates a curried function from a normal function of callable object.

The curry function is itself curried and can be partially
applied. It can be used either as a normal function call or as a
decorator. The required number_of_arguments parameter makes it
possible to curry functions which take a variable number of
arguments like the built-in 'map' function.

    Usage:

    curried_map = curry(2, map)

    @curry(2)
    def some_func(x, y, z):
        return x + y - z

Args:
    number_of_arguments: The number of arguments function_to_curry
    takes as input. If function_to_curry takes a variable number of
    arguments, then number of curried arguments desired in the
    result: function_to_curry will be called once this many
    arguments have been supplied.
    function_to_curry: The function that we wish to curry

Returns:
    A new function which may be partially applied simply by passing
    the desired number of arguments.
"""

def identity(value: T) -> T:
    """ Returns it's input value unchanged. """
    return value

def kleisli_compose(
        function_f: Callable[[R], monad.Monad[S]], function_g: Callable[[S], monad.Monad[T]]
) -> Callable[[R], monad.Monad[T]]:
    """ Composes two Kleisli functions.

    Kleisli functions are functions which take as input a 'bare' value
    and return an 'embellished' value. For instance, if we have a
    function f which maps a's to b's, it's type is:
      f :: a -> b

    Then the corresponding Kleisli function, f_kleisli has the type:
      f_kleisli :: a -> (b, m)

    The type (b, m) corresponds to the internal representation of the
    Monad class, so in terms of pymonad, a Kleisli function is one
    which maps values of type a to values of some sub-class of Monad.

      Example:
        def fail_if_zero(x):
            return Nothing if x is zero else Just(x)

        def add1(x):
            return Just(x + 1)

        new_function = kleisli_compose(add1, fail_if_zero)
        new_function(0) # returns Just(1)
        new_function(-1) # returns Nothing

    add1 and fail_if_zero are Kleisli functions and new_function is
    the function which results from first performing add1 followed by
    fail_if_zero.

    Args:
      function_f: a function with type: a -> (b, m)
      function_g: a function with type: b -> (c, m)

    Returns:
      A new Kleisli function with type: a -> (c, m)
    """
    return lambda a: function_f(a).bind(function_g)



# --------------------------------------------------------
# (c) Copyright 2014, 2020 by Jason DeLaat.
# Licensed under BSD 3-clause licence.
# --------------------------------------------------------
""" Implements the State monad. """

from typing import Any, Callable, Generic, Tuple, TypeVar, Union # pylint: disable=unused-import

import pymonad.monad
import pymonad.tools

A = TypeVar('A') # pylint: disable=invalid-name
B = TypeVar('B') # pylint: disable=invalid-name
S = TypeVar('S') # pylint: disable=invalid-name

@pymonad.tools.curry(3)
def _bind(monad_value, kleisli_function, state):
    return kleisli_function(monad_value.value(state)).run(monad_value.monoid(state))

@pymonad.tools.curry(3)
def _bind_or_map(monad_value, function, state):
    try:
        return _bind(monad_value, function, state)
    except AttributeError:
        return _map(monad_value, function, state)

@pymonad.tools.curry(3)
def _map(monad_value, function, state):
    return (function(monad_value.value(state)), monad_value.monoid(state))

class _State(pymonad.monad.Monad, Generic[S, A]):
    @classmethod
    def insert(cls, value: A) -> '_State[Any, A]':
        """ See Monad.insert. """
        return State(lambda s: (value, s))

    def amap(self: '_State[S, Callable[[A], B]]', monad_value: '_State[S, A]') -> '_State[S, B]':
        """ See Monad.amap. """
        return State(lambda s:
                     (self.value(s)(monad_value.value(s)),
                      monad_value.monoid(s)))

    def bind(
            self: '_State[S, A]', kleisli_function: Callable[[A], '_State[S, B]']
    ) -> '_State[S, B]':
        """ See Monad.bind. """
        return State(_bind(self, kleisli_function)) # pylint: disable=no-value-for-parameter

    def map(self: '_State[S, A]', function: Callable[[A], B]) -> '_State[S, B]':
        """ See Monad.map. """
        return State(_map(self, function)) # pylint: disable=no-value-for-parameter

    def run(self: '_State[S, A]', input_state: S) -> Tuple[A, S]:
        """ Gives the state calculation an initial state and computes the result.

        Args:
          input_state: the initial state for the stateful calculation

        Result:
          A tuple containing the result of the stateful calculation
          and the final state.
        """
        return self.value(input_state), self.monoid(input_state)

    def then(
            self: '_State[S, A]', function: Union[Callable[[A], B], Callable[[A], '_State[S, B]']]
    ) -> '_State[S, B]':
        return State(_bind_or_map(self, function)) # pylint: disable=no-value-for-parameter

def State(state_function: Callable[[S], A]) -> _State[S, A]: # pylint: disable=invalid-name
    """ The State monad constructor function.

    Args:
      state_function: a function with type State -> (Any, State)

    Returns:
      An instance of the State monad.
    """
    return _State(lambda s: state_function(s)[0], lambda s: state_function(s)[1])

State.apply = _State.apply
State.insert = _State.insert



# --------------------------------------------------------
# (c) Copyright 2014, 2020 by Jason DeLaat.
# Licensed under BSD 3-clause licence.
# --------------------------------------------------------
""" Adds operators to the Writer monad. """
from typing import TypeVar

import pymonad.monad
import pymonad.operators.operators
import pymonad.writer

M = TypeVar('M') # pylint: disable=invalid-name
T = TypeVar('T') # pylint: disable=invalid-name

class Writer(pymonad.operators.operators.MonadOperators, pymonad.writer.Writer[M, T]):
    """ See pymonad.operators.operators and pymonad.writer. """



# --------------------------------------------------------
# (c) Copyright 2014, 2020 by Jason DeLaat.
# Licensed under BSD 3-clause licence.
# --------------------------------------------------------
""" Defines the monad operators.

The MonadOperators class allows users to create a monad alias for
any monad class with the operators __and__, __rmul__, and
__rshift__ defined.

  Example:
    class MaybeOp(MonadOperators, Maybe):
        pass

Nothing extra needs to be implemented to get the correct behaviour
unless the monad has multiple 'constructors' such as Just and
Nothing. In this case, we need to override those constructors as well:

  Example:
    def Just(value):
        return MaybeOp(value, True)
    Nothing = MaybeOp(None, False)
"""
import pymonad.monad


class MonadOperators(pymonad.monad.MonadAlias):
    """ Defines MonadOperators class.

    MonadOperators is a MonadAlias which is used to add operators for
    map (<<), amap (&), bind (>>), and then (*) methods to Monad classes.
    """

    def __and__(self, monad_value):
        return self.amap(monad_value)

    def __rmul__(self, function):
        return self.then(function)

    def __mul__(self, function):
        return self.then(function)

    def __rshift__(self, kleisli_function):
        return self.bind(kleisli_function)

    def __rlshift__(self, function):
        return self.map(function)



# --------------------------------------------------------
# (c) Copyright 2014, 2020 by Jason DeLaat.
# Licensed under BSD 3-clause licence.
# --------------------------------------------------------
""" Implements the Writer monad.

The Writer monad is typically used to append information to a log. The
log type is often just strings but can be any type that behaves as a
monoid with a defined + (__add__) operator.

  Example:
    @curry(2)
    def add(x, y):
        return Writer(x + y, f"Called function 'add' with arguments {x} and {y}. Result: {x + y}")

    @curry(2)
    def mul(x, y):
        return Writer(x * y, f"Called function 'mul' with arguments {x} and {y}. Result: {x * y}")

    logged_arithmetic = (Writer
                         .insert(0)
                         .then(add(1))
                         .then(mul(2)))

    # logged_arithmetic = (2, "Called function 'add' with arguments 1 and 0. Result: 1
    #                     Called function 'mul' with arguments 2 and 1. Result: 2")
"""
from typing import Callable, Generic, List, TypeVar

import pymonad.monad
import pymonad.monoid

M = TypeVar('M', int, str, List, pymonad.monoid.Monoid, pymonad.monoid.ZERO) # pylint: disable=invalid-name
S = TypeVar('S') # pylint: disable=invalid-name
T = TypeVar('T') # pylint: disable=invalid-name

class Writer(pymonad.monad.Monad, Generic[M, T]):
    """ The Writer monad class. """
    @classmethod
    def insert(cls, value: T) -> 'Writer[M, T]':
        """ See Monad.insert. """
        return Writer(value, pymonad.monoid.ZERO)

    def bind(
            self: 'Writer[M, S]', kleisli_function: Callable[[S], 'Writer[M, T]']
    ) -> 'Writer[M, T]':
        """ See Monad.bind. """
        result = kleisli_function(self.value)
        return Writer(result.value, self.monoid + result.monoid)

    def map(self: 'Writer[M, S]', function: Callable[[S], T]) -> 'Writer[M, T]':
        """ See Monad.map. """
        return Writer(function(self.value), self.monoid)

    def __eq__(self, other):
        return self.value == other.value and self.monoid == other.monoid

    def __repr__(self):
        return f'({self.value}, {self.monoid})'



# --------------------------------------------------------
# (c) Copyright 2014, 2020 by Jason DeLaat.
# Licensed under BSD 3-clause licence.
# --------------------------------------------------------
""" Adds operators to the Either monad. """
from typing import Any, TypeVar

import pymonad.either
import pymonad.monad
import pymonad.operators.operators

S = TypeVar('S') # pylint: disable=invalid-name
T = TypeVar('T') # pylint: disable=invalid-name

class Either(pymonad.operators.operators.MonadOperators, pymonad.either.Either[S, T]):
    """ See pymonad.operators.operators and pymonad.either. """

def Left(value: S) -> Either[S, Any]: # pylint: disable=invalid-name
    """ Creates a value of the first possible type in the Either monad. """
    return Either(None, (value, False))

def Right(value: T) -> Either[Any, T]: # pylint: disable=invalid-name
    """ Creates a value of the second possible type in the Either monad. """
    return Either(value, (None, True))







class _Error(Either[S, T]): # pylint: disable=too-many-ancestors
    def __repr__(self):
        return f'Result: {self.value}' if self.is_right() else f'Error: {self.monoid[0]}' # pylint: disable=no-member

def Error(value: S) -> _Error[S, Any]: # pylint: disable=invalid-name
    """ Creates an error value as the result of a calculation. """
    return _Error(None, (value, False))

def Result(value: T) -> _Error[Any, T]: # pylint: disable=invalid-name
    """ Creates a value representing the successful result of a calculation. """
    return _Error(value, (None, True))

Error.insert = _Error.insert



# --------------------------------------------------------
# (c) Copyright 2014, 2020 by Jason DeLaat.
# Licensed under BSD 3-clause licence.
# --------------------------------------------------------
""" Adds operators to the Maybe monad. """
from typing import TypeVar

import pymonad.maybe
import pymonad.monad
import pymonad.operators.operators

T = TypeVar('T') # pylint: disable=invalid-name

class Maybe(pymonad.operators.operators.MonadOperators, pymonad.maybe.Maybe[T]):
    """ See pymonad.operators.operators and pymonad.maybe. """

def Just(value: T) -> Maybe[T]: # pylint: disable=invalid-name
    """ A Maybe object representing the presence of an optional value. """
    return Maybe(value, True)

# A Maybe object representing the absence of an optional value.
Nothing = Maybe(None, False) # pylint: disable=invalid-name





class Option(Maybe[T]): # pylint: disable=too-many-ancestors
    """ An alias for the Maybe monad class. """
    def __repr__(self):
        return f'Some {self.value}' if self.monoid else 'Nothing' # pylint: disable=no-member

def Some(value: T) -> Option[T]: # pylint: disable=invalid-name
    """ An Option object representing the presence of an optional value. """
    return Option(value, True)



# --------------------------------------------------------
# (c) Copyright 2014, 2020 by Jason DeLaat.
# Licensed under BSD 3-clause licence.
# --------------------------------------------------------
""" Adds operators to the Reader monad. """
from typing import Callable, TypeVar

import pymonad.monad
import pymonad.operators.operators
import pymonad.reader

R = TypeVar('R') # pylint: disable=invalid-name
T = TypeVar('T') # pylint: disable=invalid-name

class _Reader(pymonad.operators.operators.MonadOperators, pymonad.reader._Reader[R, T]): # pylint: disable=protected-access
    """ See pymonad.operators.operators and pymonad.reader. """

def Reader(function: Callable[[R], T]) -> _Reader[R, T]: # pylint: disable=invalid-name
    """ Creates an instance of the Reader monad.

    Args:
      function: a function which takes the read-only data as input and
        returns any appropriate type.

    Result:
      An instance of the Reader monad.
    """
    return _Reader(function, None)

Reader.apply = _Reader.apply
Reader.insert = _Reader.insert



# --------------------------------------------------------
# (c) Copyright 2014, 2020 by Jason DeLaat.
# Licensed under BSD 3-clause licence.
# --------------------------------------------------------
""" Implements the Monad base class.

The Monad base class is an abstract class which defines the operations
available on all monad instances. To create a new Monad instance,
users should create a class which inherits from Monad and provides
implementations for the methods map, amap, bind, and class method
insert. See the documentation for those methods for more information on
how to implement them properly.
"""

from typing import Callable, Generic, TypeVar, Union

S = TypeVar('S') # pylint: disable=invalid-name
T = TypeVar('T') # pylint: disable=invalid-name

class Monad(Generic[T]):
    """
    Represents a "context" in which calculations can be executed.

    You won't create 'Monad' instances directly. Instead, sub-classes implement
    specific contexts. Monads allow you to bind together a series of calculations
    while maintaining the context of that specific monad.

    """
    def __init__(self, value, monoid):
        """ Initializes the internal values of the monad instance.

        All monads can be expressed as a tuple, (a, m). Representing
        all monads internally in this canonical form allows for some
        interesting effects such as easily aliasing existing monads
        instances and, if desired, adding operators. Occasionally it
        also makes implementation of the monad methods itself easier.

        Args:
          value: if we think of monads as storing some data of
            interest plus some 'meta data', then 'value' is the data of
            interest. Exactly what 'value' is/means will depend on the
            specific context of the monad in question.
          monoid: this is the 'meta data' part. While implementers may
            use an instance of the Monoid class here it is not
            required. However, the value passed in here should be a type
            that can be treated as a monoid, such as integers; strings;
            lists; etc., in order to ensure that the monad laws are
            obeyed. This is not enforced but it will result in an
            incorrect implementation.
        """
        self.value = value
        self.monoid = monoid

    @classmethod
    def apply(cls, function):
        """ Supplies a cleaner interface for applicative functor/amap usage.

        Example:
          @curry(2)
          def add(a, b): return a + b

          x = Just(1)
          y = Just(2)

          (Maybe.apply(add)
                .to_arguments(x, y)
          ) # results in Just(3)

        Args:
          function: A regular function which returns non-monadic values.

        Returns:
          A monad object based on the input class with the wrapped
          function and a new method, 'to_arguments' which will apply
          the function.
        """
        class _Applicative(cls):
            """ An internal class which provides the 'to_arguments' method. """
            amap = cls.amap
            bind = cls.bind
            insert = cls.insert
            map = cls.map
            @staticmethod
            def to_arguments(*args):
                """ Applies arguments to the function wrapped by the call to the apply method.

                Args:
                  *args: a variable number of arguments to be supplied
                     to the function wrapped by a previous call to the
                     'apply method.

                Returns:
                  A monadic value of type 'cls'
                """
                result = cls.insert(function)
                for arg in args:
                    result = result.amap(arg)
                return cls(result.value, result.monoid)

        return _Applicative(None, None) # We don't actually care about the inputs here

    @classmethod
    def insert(cls, value: T) -> 'Monad[T]':
        """ Returns an instance of the Functor with 'value' in a minimum context.  """
        raise NotImplementedError

    def amap(self: 'Monad[Callable[[S], T]]', monad_value: 'Monad[S]') -> 'Monad[T]':
        """ Applies the function stored in the functor to the value inside 'functor_value'
        returning a new functor value.
        """
        return monad_value.map(self.value)

    def bind(self: 'Monad[S]', kleisli_function: Callable[[S], 'Monad[T]']) -> 'Monad[T]':
        """ Applies 'function' to the result of a previous monadic calculation. """
        raise NotImplementedError

    def map(self: 'Monad[S]', function: Callable[[S], T]) -> 'Monad[T]':
        """ Applies 'function' to the contents of the functor and returns a new functor value. """
        raise NotImplementedError("'fmap' not defined.")

    def then(
            self: 'Monad[S]', function: Union[Callable[[S], T], Callable[[S], 'Monad[T]']]
    ) -> 'Monad[T]':
        """ Combines the functionality of bind and fmap.

        Instead of worrying about whether to use bind or fmap,
        users can just use the then method to chain function
        calls together. The then method uses attempts to use
        bind first and if that doesn't work, uses fmap
        instead.

        Args:
          function: A python function or lambda expression
            which returns either a build-in type (int, string,
            etc.) or an appropriate monad type (Maybe, Either,
            etc.)

        Returns:
          A monad value of the same type as 'self'
        """
        try:
            result = self.bind(function)
            if isinstance(result, Monad): # pylint: disable=no-else-return
                return result
            else:
                return self.map(function)
        except AttributeError:
            return self.map(function)


class MonadAlias(Monad[T]):
    """ Provides monad method overrides which make it easy to give a monad an alias.

    MonadAlias provides monad methods which call their base class but
    then re-cast the result into the alias class. This gives users a
    mechanism to give monads more semantically meaningful names
    without needing to re-implement them.

      Example:
        class Option(MonadAlias, Maybe): # MonadAlias must be the first parent class
            def __repr__(self):
                return f'Some {self.value}' if self.monoid else 'Nothing'

        def Some(value):
            return Option(value, True)

        # Nothing = Nothing, we don't need to change that one.

    Functions which have been previously defined to work with the base
    type, Maybe in this case, work transparently with the aliased type
    without needing to be re-written.

      Example:
        def add_1(x): return Just(x + 1) # Written for the Maybe monad.

        opt_val = Option.insert(1).then(add_1) # Result: Some(2)
    """
    @classmethod
    def insert(cls, value):
        result = super().insert(value)
        return cls(result.value, result.monoid)

    def amap(self, monad_value):
        result = super().amap(monad_value)
        return self.__class__(result.value, result.monoid)

    def bind(self, kleisli_function):
        result = super().bind(kleisli_function)
        return self.__class__(result.value, result.monoid)

    def map(self, function):
        result = super().map(function)
        return self.__class__(result.value, result.monoid)

    def then(self, function):
        result = super().then(function)
        return self.__class__(result.value, result.monoid)



# --------------------------------------------------------
# (c) Copyright 2020 by Jason DeLaat.
# Licensed under BSD 3-clause licence.
# --------------------------------------------------------
""" Implements the Promise monad for ordering concurrent computations.

The Promise monad is based on (and named after) Javascript's Promise
objects and function in a similar way. Promises take asynchronous
computations and ensure the ordering of execution. In addition to the
standard operations on monads, Promises also provide a 'catch' method
which allows for recovery from errors.

  Example:
    import asyncio
    from pymonad.promise import Promise
    from pymonad.tools import curry

    @curry(2)
    def div(y, x):
        return x / y

    async def long_id(x):
        await asyncio.sleep(1)
        return await Promise(lambda resolve, reject: resolve(x))

    async def main():
        x = (Promise.insert(1)
             .then(long_id))
        y = (Promise
             .insert(2)
             .then(long_id)
             .then(div(0))            # Raises an error...
             .catch(lambda error: 2)) # ...which is dealth with here.
        print(
            await Promise.apply(add)
            .to_arguments(x, y)
            .catch(lambda error: 'Recovering...') # This is ignored
                                                  # because the previous
                                                  # catch already dealt
                                                  # with the error.
        )

    asyncio.run(main())

The above example will print the value '3' to the screen. The
'long_id' coroutine is a stand-in for any async operation that may
take some amount of time. When we await the Promise inside the print()
call it waits for both arguments to complete before calling 'add' with
the results. If the first call to 'catch' were removed then the error
would propagate and be caught by the second call. The program would
then print the string 'Recovering...' instead of '3'.

You can also create a Promise by passing a function directly. This
function takes two callbacks as input to signal a successful
computation or a failed one.

  Example:
    import asyncio

    def main():
        print(await Promise(lambda resolve, reject: resolve(2)))

    asyncio.run(main())

The 'resolve' callback can take a value of any type but the 'reject'
callback should always take an Exception as its argument.

  Example:
    import asyncio

    def main():
        print(await Promise(lambda resolve, reject: reject(IndexError())))

    asyncio.run(main())

When run, this program will crash having raised the IndexError without
catching it. Similarly, the catch method takes a function which
accepts an Exception as it's input.

  Example:
    import asyncio

    def main():
        print(await Promise(lambda resolve, reject: reject(IndexError()))
              .catch(lambda error: print(type(error)))
        )

    asyncio.run(main())

This program prints "<type 'IndexError'>" as its output.
"""
from typing import Callable, Generic, TypeVar, Union

import pymonad.monad
import pymonad.tools

S = TypeVar('S') # pylint: disable=invalid-name
T = TypeVar('T') # pylint: disable=invalid-name

ResolveFunction = Callable[[S], T]
RejectFunction = Callable[[Exception], T]
PromiseFunction = Callable[[ResolveFunction, RejectFunction], T]

def _reject(error):
    if not isinstance(error, Exception): # pylint: disable=no-else-raise
        raise Exception(str(error))
    else:
        raise error

class _Promise(pymonad.monad.Monad, Generic[T]):
    def __init__(self, value, monoid):
        super().__init__(value, monoid)
        self._resolve = pymonad.tools.identity

    @classmethod
    def insert(cls, value: T) -> '_Promise[T]':
        """ See Monad.insert. """
        return Promise(lambda resolve, reject: resolve(value))

    def amap(self: '_Promise[Callable[[S], T]]', monad_value: '_Promise[S]') -> '_Promise[T]':
        """ See Monad.amap. """
        async def _awaitable_amap(resolve, reject): # pylint: disable=unused-argument
            function = await self
            value = await monad_value
            return resolve(function(value))
        return self.__class__(_awaitable_amap, None)

    def bind(self: '_Promise[S]', kleisli_function: Callable[[S], '_Promise[T]']) -> '_Promise[T]':
        """ See Monad.bind. """
        self._resolve = kleisli_function
        async def _awaitable_bind(resolve, reject): # pylint: disable=unused-argument
            value = await self
            return resolve(await value)
        return self.__class__(_awaitable_bind, None)

    def catch(self: '_Promise[T]', error_handler: Callable[[Exception], T]) -> '_Promise[T]':
        """ Allows users to handle errors caused earlier in the Promise chain.

        The catch method takes an error handling function as input. If
        an earlier computation in the Promise chain has caused an
        error, either by being passed an Exception via the 'reject'
        callback or by an Exception being raised normally, then error
        handler is called with the Exception as an argument. If no
        error was previously raised, the error handler is ignored.

        Arguments:
          error_handler: a function which takes an Exception as input
            and can return any type.

        Returns:
          A new Promise object.
        """
        async def _awaitable_catch(resolve, reject): # pylint: disable=unused-argument
            try:
                value = await self
                return resolve(value)
            except Exception as e: # pylint: disable=invalid-name, broad-except
                return resolve(error_handler(e))

        return self.__class__(_awaitable_catch, None)

    def map(self: '_Promise[S]', function: Callable[[S], T]) -> '_Promise[T]':
        """ See Monad.map. """
        self._resolve = function
        async def _(resolve, reject): # pylint: disable=unused-argument
            value = await self
            return resolve(value)
        return self.__class__(_, None)


    def then(
            self: '_Promise[S]', function: Union[Callable[[S], T], Callable[[S], '_Promise[T]']]
    ) -> '_Promise[T]':
        """ See Monad.then. """
        async def _awaitable_then(resolve, reject): # pylint: disable=unused-argument
            try:
                return resolve(await self.bind(function))
            except TypeError:
                return resolve(await self.map(function))
        return self.__class__(_awaitable_then, None)

    def __await__(self):
        return self.value(self._resolve, _reject).__await__()

def Promise(function: PromiseFunction) -> _Promise[T]: # pylint: disable=invalid-name
    """ Constructs a Promise object for ordering concurrent computations.

    Example:
      Promise(lambda resolve, reject: resolve('any value'))

      def some_computation(resolve, reject):
          if True:
              return resolve(10)
          else:
              reject(TypeError('Fake error.')) # doesn't need to be returned
      Promise(some_computation)

    Arguments:
      function: a function taking two callback typically called
        'resolve' and 'reject'. When the computation is successful the
        value should be returned by calling resolve with the result. If
        there is an error, call 'reject' with an instance of the
        Exception class.

    Returns:
      A new Promise object.
    """
    @pymonad.tools.curry(3)
    async def _awaitable(function, resolve, reject):
        return function(resolve, reject)
    return _Promise(_awaitable(function), None) # pylint: disable=no-value-for-parameter

Promise.apply = _Promise.apply
Promise.insert = _Promise.insert



# --------------------------------------------------------
# (c) Copyright 2014, 2020 by Jason DeLaat.
# Licensed under BSD 3-clause licence.
# --------------------------------------------------------
""" Implements the Maybe monad and related functions.

The Maybe monad is used to represent calculations that may or may not
return a value. Alternately, if used as function inputs, Maybe values
can be used to indicate 'optional' inputs, explicitly passing
'Nothing' when no input is required.

When creating Maybe values directly use the 'Just' function or 'Nothing':

  Example:
    x = Just(19)
    y = Just('A string')
    z = Nothing

The 'insert' class method is a wrapper around the 'Just' function.

  Example:
    x = Maybe.insert(9) # Same as Just(9)
"""
from typing import Any, Callable, Generic, TypeVar

import pymonad.monad

S = TypeVar('S') # pylint: disable=invalid-name
T = TypeVar('T') # pylint: disable=invalid-name

class Maybe(pymonad.monad.Monad, Generic[T]):
    """ The Maybe monad class. """
    @classmethod
    def insert(cls, value: T) -> 'Maybe[T]':
        """ See Monad.insert """
        return Just(value)

    def amap(self: 'Maybe[Callable[[S], T]]', monad_value: 'Maybe[S]') -> 'Maybe[T]':
        """ See Monad.amap"""
        if self.is_nothing() or monad_value.is_nothing(): #pylint: disable=no-else-return
            return Nothing
        else:
            return monad_value.map(self.value)

    def bind(self: 'Maybe[S]', kleisli_function: 'Callable[[S], Maybe[T]]') -> 'Maybe[T]':
        """ See Monad.bind """
        if self.monoid is False: #pylint: disable=no-else-return
            return self
        else:
            try:
                return kleisli_function(self.value)
            except: # pylint: disable=bare-except
                return Nothing

    def is_just(self) -> bool:
        """ Returns True if the monad instance was created with the 'Just' function. """
        return self.monoid

    def is_nothing(self) -> bool:
        """ Returns True if the monad instance is the 'Nothing' value. """
        return not self.monoid

    def map(self: 'Maybe[S]', function: Callable[[S], T]) -> 'Maybe[T]':
        """ See Monad.map """
        if self.is_nothing(): #pylint: disable=no-else-return
            return self
        else:
            try:
                return Just(function(self.value))
            except: # pylint: disable=bare-except
                return Nothing

    def __eq__(self, other):
        """ Checks equality of Maybe objects.

        Maybe objects are equal iff:
          1. They are both Nothing, or
          2. They are both Just and
            2a. They both contain the same value.
        """
        return self.value == other.value and self.monoid == other.monoid

    def __repr__(self):
        return f'Just {self.value}' if self.monoid else 'Nothing'

def Just(value: T) -> Maybe[T]: # pylint: disable=invalid-name
    """ A Maybe object representing the presence of an optional value. """
    return Maybe(value, True)

# A Maybe object representing the absence of an optional value.
Nothing: Maybe[Any] = Maybe(None, False) # pylint: disable=invalid-name







class Option(pymonad.monad.MonadAlias, Maybe[T]): # MonadAlias must be the first parent class
    """ An alias for the Maybe monad class. """
    def __repr__(self):
        return f'Some {self.value}' if self.monoid else 'Nothing'

def Some(value: T) -> Option[T]: # pylint: disable=invalid-name
    """ An Option object representing the presence of an optional value. """
    return Option(value, True)



# --------------------------------------------------------
# (c) Copyright 2014, 2020 by Jason DeLaat.
# Licensed under BSD 3-clause licence.
# --------------------------------------------------------
""" Implements the Either monad and related functions.

The Either type represents values that can either type A or type B -
for any types A and B - but not both at the same time. As a function
input type, Either values can be used to define functions which can
sensibly deal with multiple types of input; of course in python we
don't need a special way to deal with multiple input types.

Perhaps more usefully, as an output type, Either values can be used to
signal that a function may cause an error: Either we get back a useful
result or we get back an error message.

When creating Either values directly use the 'Right' or 'Left'
functions:

  Example:
    x = Right(19)                     # Represents a result value
    y = Left('Something went wrong.') # Represents an error value

The 'insert' class method is a wrapper around the 'Right' function.

  Example:
    x = Either.insert(9) # Same as Right(9)
"""
from typing import Any, Callable, Generic, TypeVar

import pymonad.monad

M = TypeVar('M') # pylint: disable=invalid-name
S = TypeVar('S') # pylint: disable=invalid-name
T = TypeVar('T') # pylint: disable=invalid-name

class Either(pymonad.monad.Monad, Generic[M, T]):
    """ The Either monad class. """
    @classmethod
    def insert(cls, value: T) -> 'Either[Any, T]':
        """ See Monad.insert """
        return Right(value)

    def amap(self: 'Either[M, Callable[[S], T]]', monad_value: 'Either[M, S]') -> 'Either[M, T]':
        if self.is_left(): # pylint: disable=no-else-return
            return self
        elif monad_value.is_left(): # pylint: disable=no-else-return
            return monad_value
        else:
            return Right(self.value(monad_value.value))

    def bind(
            self: 'Either[M, S]', kleisli_function: Callable[[S], 'Either[M, T]']
    ) -> 'Either[M, T]':
        """ See Monad.bind """
        if self.is_left(): # pylint: disable=no-else-return
            return self
        else:
            try:
                return kleisli_function(self.value)
            except Exception as e: # pylint: disable=invalid-name, broad-except
                return Left(e)

    def is_left(self) -> bool:
        """ Returns True if this Either instance was created with the 'Left' function. """
        return not self.monoid[1]

    def is_right(self) -> bool:
        """ Returns True if this Either instance was created with the 'Right' function. """
        return self.monoid[1]

    def map(self: 'Either[M, S]', function: Callable[[S], T]) -> 'Either[M, T]':
        """ See Monad.map """
        if self.is_left(): # pylint: disable=no-else-return
            return self
        else:
            try:
                return Right(function(self.value))
            except Exception as e: # pylint: disable=invalid-name, broad-except
                return Left(e)

    def __eq__(self, other):
        """ Checks equality of Maybe objects.

        Maybe objects are equal iff:
          1. They are both Nothing, or
          2. They are both Just and
            2a. They both contain the same value.
        """
        return self.value == other.value and self.monoid == other.monoid

    def __repr__(self):
        return f'Right {self.value}' if self.is_right() else f'Left {self.monoid[0]}'

def Left(value: M) -> Either[M, Any]: # pylint: disable=invalid-name
    """ Creates a value of the first possible type in the Either monad. """
    return Either(None, (value, False))

def Right(value: T) -> Either[Any, T]: # pylint: disable=invalid-name
    """ Creates a value of the second possible type in the Either monad. """
    return Either(value, (None, True))






class _Error(pymonad.monad.MonadAlias, Either[M, T]):
    def __repr__(self):
        return f'Result: {self.value}' if self.is_right() else f'Error: {self.monoid[0]}'

def Error(value: M) -> _Error[M, Any]: # pylint: disable=invalid-name
    """ Creates an error value as the result of a calculation. """
    return _Error(None, (value, False))

def Result(value: T) -> _Error[Any, T]: # pylint: disable=invalid-name
    """ Creates a value representing the successful result of a calculation. """
    return _Error(value, (None, True))

Error.apply = _Error.apply
Error.insert = _Error.insert



# --------------------------------------------------------
# (c) Copyright 2014, 2020 by Jason DeLaat.
# Licensed under BSD 3-clause licence.
# --------------------------------------------------------
""" Adds operators to the State monad. """
from typing import Callable, Tuple, TypeVar

import pymonad.monad
import pymonad.operators.operators
import pymonad.state

S = TypeVar('S') # pylint: disable=invalid-name
T = TypeVar('T') # pylint: disable=invalid-name

class _State(pymonad.operators.operators.MonadOperators, pymonad.state._State[S, T]): # pylint: disable=protected-access
    """ See pymonad.operators.operators and pymonad.state. """

def State(state_function: Callable[[S], Tuple[T, S]]) -> _State[S, T]: # pylint: disable=invalid-name
    """ The State monad constructor function.

    Args:
      state_function: a function with type State -> (Any, State)

    Returns:
      An instance of the State monad.
    """
    return _State(lambda s: state_function(s)[0], lambda s: state_function(s)[1])

State.apply = _State.apply
State.insert = _State.insert



# --------------------------------------------------------
# (c) Copyright 2014, 2020 by Jason DeLaat.
# Licensed under BSD 3-clause licence.
# --------------------------------------------------------
""" Implements the base Monoid type.

A monoid is an algebraic structure consisting of a set of objects, S,
such as integers; strings; etc., and an operation usually denoted as
'+' which obeys the following rules:
  1. Closure: If 'a' and 'b' are in S, then 'a + b' is also in S.
  2. Identity: There exists an element in S (denoted 0) such that
     a + 0 = 0 + a = a
  3. Associativity: (a + b) + c = a + (b + c)

For monoid types, the '+' operation is implemented by the method
'mplus' and the static method 'mzero' is defined to return the
identity element of the type.

For example, integers can be monoids in two ways:
  1. mzero = 0 and mplus = addition
  2. mzero = 1 and mplus = multiplication

String can also form a monoid where mzero is the empty string and
mplus is concatenation.
"""

from typing import Any, Generic, List, TypeVar # pylint: disable=unused-import

T = TypeVar('T') # pylint: disable=invalid-name

class _MonoidZeroMeta(type):
    def __add__(cls, other):
        return other

    def __radd__(cls, other):
        return other

    def __repr__(cls):
        return 'MZERO'

class ZERO(metaclass=_MonoidZeroMeta): # pylint: disable=too-few-public-methods
    """ A generic zero/identity element for monoids.

    The ZERO class acts as a constant/singleton with monoid addition
    implemented on the class itself to always return the other
    element. It is not actually possible to create an instance of ZERO
    as calling the constructor simply returns the class itself.

    Example:
      ZERO == ZERO() # True.
      ZERO + 10      # 10
      'hello' + ZERO # 'hello'
    """
    def __new__(cls):
        return ZERO

class Monoid(Generic[T]):
    """ Abstract base class for Monoid instances.

    To implement a monoid instance, users should create a sub-class of
    Monoid and implement the mzero and mplus methods. Additionally, it
    is the implementers responsibility to ensure that the
    implementation adheres to the closure, identity and associativity
    laws for monoids.
    """

    def __init__(self, value: T) -> None:
        """ Initializes the monoid element to 'value'.  """
        self.value = value

    def __add__(self: 'Monoid[T]', other: 'Monoid[T]') -> 'Monoid[T]':
        """ The 'mplus' operator.  """
        return self.mplus(other)

    def __eq__(self: 'Monoid[T]', other: 'Monoid[T]') -> bool:
        return self.value == other.value


    @staticmethod
    def mzero() -> 'Monoid[Any]':
        """
        A static method which simply returns the identity value for the monoid type.
        This method must be overridden in subclasses to create custom monoids.
        See also: the mzero function.

        """
        raise NotImplementedError

    def mplus(self, other):
        """
        The defining operation of the monoid. This method must be overridden in subclasses
        and should meet the following conditions.
           1. x + 0 == 0 + x == x
           2. (x + y) + z == x + (y + z) == x + y + z
        Where 'x', 'y', and 'z' are monoid values, '0' is the mzero (the identity value) and '+'
        is mplus.

        """
        raise NotImplementedError

def mzero(monoid_type):
    """
    Returns the identity value for monoid_type. Raises TypeError if
    monoid_type is not a valid monoid.

    There are a number of builtin types that can operate as monoids
    and they can be used as such as is. These "natural" monoids are:
    int, float, str, and list.  While thee mzero method will work on
    monoids derived from the Monoid class, this mzero function will
    work for *all* monoid types, including the "natural" monoids.  For
    this reason it is preferable to call this function rather than
    calling the mzero method directly unless you know for sure what
    type of monoid you're dealing with.

    """
    try:
        return monoid_type.mzero()
    except AttributeError:
        if (isinstance(monoid_type, (int, float)) or monoid_type == int or monoid_type == float): # pylint: disable=no-else-return
            return 0
        elif isinstance(monoid_type, str) or monoid_type == str:
            return ""
        elif isinstance(monoid_type, list) or monoid_type == list:
            return []
        else:
            raise TypeError(str(monoid_type) + " is not a Monoid.")

def mconcat(monoid_list: List[Monoid[T]]) -> Monoid[T]:
    """
    Takes a list of monoid values and reduces them to a single value by applying the
    mplus operation to each all elements of the list.

    """
    result = mzero(monoid_list[0])
    for value in monoid_list:
        result += value
    return result



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pymonad', 'pymonad.operators']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'boxx-pymonad',
    'version': '0.2.0',
    'description': 'My Implementation of pymonad package.',
    'long_description': '============\nboxx-pymonad\n============\n\n\n.. image:: https://img.shields.io/pypi/v/boxx_pymonad.svg\n        :target: https://pypi.python.org/pypi/boxx_pymonad\n\n.. image:: https://img.shields.io/travis/wboxx1/boxx-pymonad.svg\n        :target: https://travis-ci.org/wboxx1/boxx-pymonad\n\n.. image:: https://ci.appveyor.com/api/projects/status/wboxx1/branch/master?svg=true\n    :target: https://ci.appveyor.com/project/wboxx1/boxx-pymonad/branch/master\n    :alt: Build status on Appveyor\n\n.. image:: https://readthedocs.org/projects/boxx-pymonad/badge/?version=latest\n        :target: https://boxx-pymonad.readthedocs.io/en/latest/?badge=latest\n        :alt: Documentation Status\n\n\n\n\nMy Implementation of pymonad package.\n\n\n* Free software: MIT license\n\n* Documentation: https://wboxx1.github.io/boxx-pymonad\n\n\n\nInstallation:\n-------------\n\n.. code-block:: console\n\n    $ pip install boxx-pymonad\n\nFeatures\n--------\n\n* TODO\n\nCredits\n-------\n\nThis package was created with Cookiecutter_ and the `wboxx1/cookiecutter-pypackage-poetry`_ project template.\n\n.. _Cookiecutter: https://github.com/audreyr/cookiecutter\n.. _`wboxx1/cookiecutter-pypackage-poetry`: https://github.com/wboxx1/cookiecutter-pypackage-poetry\n',
    'author': 'Will Boxx',
    'author_email': 'wboxx1@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://wboxx1.github.io/boxx-pymonad',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
}


setup(**setup_kwargs)
